#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations

import dataclasses as dc
import enum
from collections.abc import Mapping
from typing import Any, Dict, Sequence, Union

import click

_LOG_VERBOSITY = 0


@enum.unique
class AddressOffset(enum.IntEnum):
    """
    Enumeration of address bit offsets, defined by Address Format of the product specification.
    """

    REGION = 29
    SECURITY = 28
    DOMAINID = 24
    ADDR = 23


@enum.unique
class Product(enum.Enum):
    """
    Enumeration of supported product names in the Haltium family.
    """

    NRF54H20 = enum.auto()
    NRF9280 = enum.auto()

    @classmethod
    def _missing_(cls, value: str):
        """
        Custom enum behavior to support case insensitive values.
        """
        return cls.__members__.get(value.upper())


class _HaltiumID(enum.Enum):
    """
    Common parent class of Haltium ID enums.
    """

    def __int__(self):
        """Integer representation of the class."""
        return int(self.value)

    def __lt__(self, other):
        """Overridden less-than to guarantee check of Enum int values."""
        if self.value < int(other):
            return True
        return False

    def __str__(self):
        """String representation of the class."""
        return self.name

    @classmethod
    def from_value(cls, value: int):
        """
        Helper method to return the enum object equivalent to an integer, if it is enumerated.
        """
        for x in cls:
            if int(x) == value:
                return x
        return None

    def id_string(self):
        return f"{self.value} ({self.name})"


@enum.unique
class AddressRegion(_HaltiumID):
    """
    Enumeration of address regions, defined by Address Format of the product specification.
    """

    PROGRAM = 0
    DATA = 1
    PERIPHERAL = 2
    EXT_XIP = 3
    EXT_XIP_ENCRYPTED = 4
    STM = 5
    CPU = 7

    @classmethod
    def from_address(cls, address: int) -> AddressRegion:
        """
        Helper method to extract an enumerated address region from a given address.
        """
        return cls.from_value((address >> AddressOffset.REGION) & 0b111)


@enum.unique
class DomainID(_HaltiumID):
    """
    Enumeration of domain IDs in Haltium products.
    """

    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    GLOBAL_FAST = 12
    GLOBAL_SLOW = 13
    GLOBAL_ = 14
    GLOBAL = 15

    @classmethod
    def from_address(cls, address: int):
        """
        Helper method to extract an enumerated domain ID from a given address.
        """
        return cls.from_value((address >> AddressOffset.DOMAINID) & 0xF)

    @classmethod
    def from_processor(cls, processor: Union[ProcessorID, int]):
        """
        Helper method to extract an enumerated domain ID from a processor ID.
        """
        processor_domain = {
            ProcessorID.SECURE.value: cls.SECURE,
            ProcessorID.APPLICATION.value: cls.APPLICATION,
            ProcessorID.RADIOCORE.value: cls.RADIOCORE,
            ProcessorID.CELLCORE.value: cls.CELLCORE,
            ProcessorID.CELLDSP.value: cls.CELLDSP,
            ProcessorID.CELLRF.value: cls.CELLRF,
            ProcessorID.ISIMCORE.value: cls.ISIMCORE,
            ProcessorID.SYSCTRL.value: cls.GLOBAL_FAST,
            ProcessorID.PPR.value: cls.GLOBAL_SLOW,
            ProcessorID.FLPR.value: cls.GLOBAL_,
        }
        return processor_domain.get(int(processor))


@enum.unique
class OwnerID(_HaltiumID):
    """
    Enumeration of ownership IDs in haltium products.
    """

    NONE = 0
    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELL = 4
    ISIMCORE = 5
    SYSCTRL = 8

    @classmethod
    def from_domain(cls, domain: DomainID):
        """
        Helper method to extract an enumerated owner ID from a domain ID.
        """
        domain_owner = {
            DomainID.SECURE.value: cls.SECURE,
            DomainID.APPLICATION.value: cls.APPLICATION,
            DomainID.RADIOCORE.value: cls.RADIOCORE,
            DomainID.CELLCORE.value: cls.CELL,
            DomainID.CELLDSP.value: cls.CELL,
            DomainID.CELLRF.value: cls.CELL,
            DomainID.ISIMCORE.value: cls.ISIMCORE,
            DomainID.GLOBAL_FAST.value: cls.SYSCTRL,
        }
        return domain_owner.get(int(domain))

    @classmethod
    def from_processor(cls, processor: Union[ProcessorID, int]):
        """
        Helper method to extract an enumerated owner ID from a processor ID.
        """
        if (domain := DomainID.from_processor(processor)) is not None:
            return cls.from_domain(domain)
        return None


@enum.unique
class ProcessorID(_HaltiumID):
    """
    Enumeration of processor IDs in haltium products.
    """

    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    BBPR = 11
    SYSCTRL = 12
    PPR = 13
    FLPR = 14

    @classmethod
    def from_domain(cls, domain: DomainID):
        """
        Helper method to extract an enumerated processor ID from a domain ID.
        """
        domain_processor = {
            DomainID.SECURE.value: cls.SECURE,
            DomainID.APPLICATION.value: cls.APPLICATION,
            DomainID.RADIOCORE.value: cls.RADIOCORE,
            DomainID.CELLCORE.value: cls.CELLCORE,
            DomainID.CELLDSP.value: cls.CELLDSP,
            DomainID.CELLRF.value: cls.CELLRF,
            DomainID.ISIMCORE.value: cls.ISIMCORE,
            DomainID.GLOBAL_FAST.value: cls.SYSCTRL,
            DomainID.GLOBAL_SLOW.value: cls.PPR,
            DomainID.GLOBAL_.value: cls.FLPR,
        }
        return domain_processor.get(int(domain))

    @classmethod
    def from_nodelabels(cls, labels: List[str]):
        """
        Helper method to extract an enumerated processor ID from a list of devicetree nodelabels.
        """
        substring_processor = {
            "cpusec": cls.SECURE,
            "cpuapp": cls.APPLICATION,
            "cpurad": cls.RADIOCORE,
            "cpucell": cls.CELLCORE,
            "cpubbpr": cls.BBPR,
            "cpusys": cls.SYSCTRL,
            "cpuppr": cls.PPR,
            "cpuflpr": cls.FLPR,
        }
        processors = {
            processor_id
            for substring, processor_id in substring_processor.items()
            if any(substring in label for label in labels)
        }
        if len(processors) == 1:
            return processors.pop()
        return None


# Bit position of the function bits in the pinctrl pin value encoded from NRF_PSEL()
NRF_PSEL_FUN_POS = 24
# Mask for the function bits in the pinctrl pin value encoded from NRF_PSEL()
NRF_PSEL_FUN_MASK = 0xFF << NRF_PSEL_FUN_POS
# Pull this from DT
GPIO_PIN_COUNT = 32


@dc.dataclass(frozen=True)
class NrfPsel:
    """Decoded NRF_PSEL values."""

    fun: int
    port: int
    pin: int


def nrf_psel_decode(psel_value: int) -> NrfPsel:
    """
    Decode an NRF_PSEL encoded GPIO pin value to its individual parts.

    :param psel_value: GPIO pin value encoded with NRF_PSEL()
    :return: A decoded tuple of GPIO (port, pin)
    """
    port, pin = gpio_port_pin_decode(psel_value & (~NRF_PSEL_FUN_MASK))
    fun = (psel_value & NRF_PSEL_FUN_MASK) >> NRF_PSEL_FUN_POS
    return NrfPsel(fun=fun, port=port, pin=pin)


def gpio_port_pin_decode(port_pin_val: int) -> Tuple[int, int]:
    """
    Decode a GPIO pin value to a tuple consisting of (port, pin)

    :param port_pin_val: GPIO pin value containing an encoded port and pin number
    :return: A decoded tuple of GPIO (port, pin)
    """
    return (port_pin_val // GPIO_PIN_COUNT, port_pin_val % GPIO_PIN_COUNT)


class Record(Dict[str, Dict[str, Union[str, int]]]):
    """
    A transcript of the contents of any number of registers and their fields.
    A basic Record is one that is not complete with register instance information, while a complete
    Record is one that also carries instance information.

    Basic example: "mem_config"
    Complete example: "mem_config_0"

    Example Format:
    record = {
        "mem_config_0": {
            "READ": "Allowed",
            "WRITE": "Allowed",
            ...
        }
        ...
    }
    """

    def update(self, other=(), /, **kwargs):
        """
        Overridden method. If a Record is given as the parameter and already has an entry in the
        mapping, update the Record contents instead. This avoids overwriting the Record entirely
        when it should instead be patched or updated as the script processes all input contents.
        If no matching Record exists or the argument is not a Record object, the parent method
        functionality is used.

        :param other: Positional arguments. See `dict.update` for more details.
        :param kwargs: Keyword arguments. See `dict.update` for more details.
        """

        if isinstance(other, Mapping):
            for key, value in other.items():
                if isinstance(self.get(key), Mapping) and isinstance(value, Mapping):
                    self[key].update(value)
                else:
                    self[key] = value
        else:
            super().update(other)

        for key, value in kwargs.items():
            if isinstance(self.get(key), Mapping) and isinstance(value, Mapping):
                self[key].update(value)
            else:
                self[key] = value

    def as_hex(self):
        """
        Helper to replace int members with their value as a hex string.
        """

        def int_to_hex(root: dict) -> Record:
            """Internal helper function for traversing the nested maps (recursive)."""

            clone = {}
            for k, v in root.items():
                if isinstance(v, int):
                    clone[k] = hex(v)
                elif isinstance(v, dict):
                    clone[k] = int_to_hex(v)
                else:
                    clone[k] = v
            return clone

        return int_to_hex(self)


def config_log(**kwargs):
    """
    Configure logging settings.

    :kwarg verbosity: Set level of verbose output
    """
    verbosity = kwargs.get("verbosity")

    if verbosity is not None:
        global _LOG_VERBOSITY
        _LOG_VERBOSITY = int(verbosity)


def log_inf(msg: str):
    """
    Print information.

    :param msg: Message to print
    """
    if _LOG_VERBOSITY >= 1:
        click.echo(msg)


def log_vrb(msg: str):
    """
    Print verbose information.

    :param msg: Message to print
    """
    if verbose_logs():
        click.secho(msg, fg="green")


def log_dbg(msg: str):
    """
    Print debug information.

    :param msg: Message to print
    """
    if _LOG_VERBOSITY >= 2:
        click.secho(msg, fg="blue")


def log_dev(msg: str):
    """
    Print detailed information for development.

    :param msg: Message to print
    """
    if _LOG_VERBOSITY >= 3:
        click.secho(msg, fg="magenta")


def log_err(msg: str):
    """
    Print error information.

    :param msg: Message to print
    """
    click.secho(f"ERROR: {msg}", err=True, fg="red")


def verbose_logs():
    """
    :return: True if verbosity is at or above the Verbose level.
    """
    return _LOG_VERBOSITY >= 1


def else_none(obj: Any, text=None) -> Union[Any, str]:
    """
    Helper function for printing an empty object nicely instead of an empty or None.

    :param obj: Object to check against being None or empty
    :kwarg text: Specific text to return

    :return: Pretty help text
    """

    if text is None and isinstance(obj, Sequence):
        text = "Empty"

    return obj if obj else str(text)



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations

from collections import ChainMap, Counter
from functools import cached_property
from itertools import chain
from pathlib import Path
from pprint import pformat
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import svd
import tomli
from intelhex import IntelHex
from svd import Device, EPath, Field, Peripheral, Register

from . import bicr, uicr
from .common import Record, log_dbg, log_vrb, log_dev
from .parsed_dt import PeripheralResult

# TOML format constants
TOML_FIELD_AS_HEX_WHEN_LARGER_THAN = 32
TOML_FIELD_COMMENT_START_AT_COLUMN = 30
TOML_FIELD_DEFAULT_VALUE_START_AT = 56


# Registers from supported peripherals that must be set as unconstrained
# regardless of commandline options.
REQUIRED_UNCONSTRAINED = {
    "BICR": bicr.BICR_UNCONSTRAINED,
}

# A map of values for a given field that are conflicting between different versions of the device
CONFLICTING_FIELD_VALUES = {
    "BICR": {
        "MODE": {
            # 'Pierce' and 'Crystal' are both used for '0' for this field
            "Pierce": 0,
            "Crystal": 0,
        }
    }
}

# Peripherals that reside in NVM
NVM_PERIPHERALS = ["UICR", "BICR"]
NVM_FILL_BYTE = 0xFF
NVM_RESET_VALUE = 0xFFFF_FFFF


class LogicalPeripheral:
    """Abstraction for a set of peripherals that are logically grouped together."""

    def __init__(
        self,
        name: str,
        device: Device,
        record: Record,
        unconstrained: Optional[List[str]] = None,
        base_address: Optional[int] = None,
        mem_map: Optional[MemoryMap] = None,
    ):
        """
        :param name: Name of the peripheral.
        :param device: SVD device element
        :param record: Peripheral register record containing register values
        :param unconstrained: List of register paths that should not be constrained
        :param base_address: Base address of the peripheral
        :param mem_map: The memory map containing register values
        """

        self._name: str = name
        self._device: Device = device

        main_peripheral = _find_peripheral(self._device, self._name)
        if base_address is not None:
            main_peripheral = main_peripheral.copy_to(base_address)

        self._peripherals: List[Peripheral] = [main_peripheral]

        if unconstrained is not None:
            self._unconstrain(unconstrained)

        if mem_map is not None:
            self._fill_from_mem_map(mem_map)

        self._fill_from_record(record)

    def _unconstrain(self, unconstrained: List[str]):
        """
        Remove restrictions put on a value held by the register.

        :param unconstrained: List of register paths that should not be constrained
        :raises ValueError: If a register is not found in the peripheral set
        """
        for register_path in unconstrained:
            record_name, field_name = split_register_path(register_path)
            try:
                register = self.record_name_to_register[record_name]
                if field_name is None:
                    register.unconstrain()
                else:
                    register[field_name].unconstrain()
            except KeyError as e:
                raise ValueError(
                    f"Unconstrained register path {register_path} not found in peripheral set "
                    f"{[p.name for p in self._peripherals]}"
                ) from e

    def _fill_from_record(self, record: Record):
        """
        Fill the values of the peripheral set based on the contents of an instance of Record.

        :param record: The record that register values are filled from
        :raises ValueError: If the register cannot be found in the peripheral set
        """
        for record_name, record_fields in record.items():
            try:
                register = self.record_name_to_register[record_name]
            except KeyError as e:
                if self._name == "UICR" and "_instance" in record_name:
                    log_dbg(f"Skipping register {record_name}")
                    continue
                else:
                    raise ValueError(
                        f"Register with record name '{record_name}' not found in peripheral set "
                        f"{[p.name for p in self._peripherals]}"
                    ) from e

            for field, value in record_fields.items():
                try:
                    register[field] = value
                except ValueError as e:
                    # The value is not found in the SVD. This could be the result of a register
                    # value enumeration having different names in different versions of the
                    # product. Check if the field in question is one of those fields, and load
                    # the raw value associated with that enumeration.
                    if (
                        self._name in CONFLICTING_FIELD_VALUES
                        and field in CONFLICTING_FIELD_VALUES[self._name]
                        and value in CONFLICTING_FIELD_VALUES[self._name][field]
                    ):
                        new_value = CONFLICTING_FIELD_VALUES[self._name][field][value]
                        log_dbg(
                            f"Setting {record_name}.{field} to {new_value} as the value '{value}'"
                            f" it is an aliased enumeration: {CONFLICTING_FIELD_VALUES}"
                        )
                        register[field] = new_value
                    else:
                        raise e

    def _fill_from_mem_map(self, mem_map: MemoryMap):
        """
        Fill the values of the peripheral set based on the contents of an instance of MemoryMap.

        :param mem_map: The memory map that register values are filled from
        :raises ValueError: If the register cannot be found in the peripheral set
        """
        mem_map_remainder: Set[int] = set(mem_map)

        for register in self.address_to_register.values():
            value, mask = mem_map[register.address_range]
            if register.content & mask != value:
                register.set_content(value, mask)

            mem_map_remainder.difference_update(register.address_range)

        if mem_map_remainder:
            raise ValueError(
                f"Address {hex(min(mem_map_remainder))} not found in peripheral set "
                f"{[p.name for p in self._peripherals]}"
            )

    @property
    def name(self) -> str:
        """Name of the peripheral set."""
        return self._name

    @cached_property
    def address_to_register(self) -> Mapping[int, Register]:
        """Map of all registers in the peripheral set, keyed by address."""
        return ChainMap(
            *(
                {
                    register.address: register
                    for register in peripheral.register_iter(leaf_only=True)
                }
                for peripheral in self._peripherals
            )
        )

    @cached_property
    def record_name_to_register(self) -> Mapping[str, Register]:
        """Map from record name to register instance."""
        return ChainMap(
            *(
                _make_record_to_register_map(peripheral)
                for peripheral in self._peripherals
            )
        )

    @cached_property
    def path_to_record_name(self) -> Mapping[EPath, str]:
        """Map of register path to record name."""
        return {
            register.path: record_name
            for record_name, register in self.record_name_to_register.items()
        }

    @cached_property
    def memory_map(self) -> Mapping[int, int]:
        """
        Return a combined memory map of the peripherals in the set.
        The returned map has a granularity of 1 byte.
        """
        return ChainMap(
            *(dict(p.memory_iter(absolute_addresses=True)) for p in self._peripherals)
        )

    @cached_property
    def written_memory_map(self) -> Mapping[int, int]:
        """
        Return a combined memory map of the peripherals in the set, but only containing those
        entries that were explicitly written to.
        The returned map has a granularity of 1 byte.
        """
        return ChainMap(
            *(
                dict(p.memory_iter(absolute_addresses=True, written_only=True))
                for p in self._peripherals
            )
        )

    @property
    def address_ranges(self) -> List[range]:
        """Return a list of address ranges covered by each peripheral in the peripheral set."""
        return [range(*peripheral.address_bounds) for peripheral in self._peripherals]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self._peripherals}"

    def __str__(self):
        return self.to_str(lambda reg: reg.modified)

    def to_str(self, reg_filter: Optional[Callable[[Register], bool]] = None):
        lines = []

        lines.append("{")

        for peripheral in self._peripherals:
            lines.append(f"  {peripheral.name} @ 0x{peripheral.base_address:08x}")

            for reg in peripheral.register_iter(leaf_only=True):
                if reg_filter is None or reg_filter(reg):
                    lines.append(f"    {reg}")

        lines.append("}")

        return "\n".join(lines)


def split_register_path(register_path: str) -> Tuple[str, Optional[str]]:
    """Split a register path such as 'register_name.field_name' into its components"""
    path_components = register_path.strip(".").rsplit(".", maxsplit=1)
    register_name = path_components[0]
    field_name = path_components[1] if len(path_components) > 1 else None
    return register_name, field_name


def _make_record_to_register_map(peripheral: Peripheral) -> Dict[str, Register]:
    """
    Make a map from record name to register based on the registers in the given peripheral.
    It is assumed that each register name maps to a unique record name in the peripheral.
    """

    register_counts: Counter = Counter()
    record_to_register_map = {}

    for register in peripheral.register_iter(leaf_only=True):
        non_indexed_path = register.path.to_flat()
        index = register_counts[non_indexed_path]
        register_counts[non_indexed_path] += 1
        record_name = f"{'_'.join(non_indexed_path).lower()}_{index}"
        record_to_register_map[record_name] = register

    return record_to_register_map


def _find_peripheral(device: Device, name: str) -> Peripheral:
    """
    Find a peripheral with the given name in the device.
    Certain common prefixes and suffixes used for peripheral names can be omitted from the name.
    """

    matches = [
        periph
        for n, periph in device.items()
        if name in (n, _strip_prefixes_suffixes(n, ["GLOBAL_"], ["_NS", "_S"]))
    ]

    if not matches:
        raise LookupError(
            f"No peripheral with name containing '{name}' found in the SVD file"
        )

    elif len(matches) > 1:
        raise ValueError(
            f"More than one peripheral with name containing '{name}' found in the "
            f"SVD file: {matches}"
        )

    return matches[0]


def _strip_prefixes_suffixes(
    word: str, prefixes: List[str], suffixes: List[str]
) -> str:
    """
    Remove a prefix and suffix from the given string.
    Up to one prefix and/or suffix is removed - if multiple of the provided strings match then
    the first found is removed.

    :param word: String to strip prefixes and suffixes from.
    :param prefixes: List of prefixes to strip.
    :param suffixes: List of suffixes to strip.

    :return: String where prefixes and suffixes have been removed.
    """

    for prefix in prefixes:
        if word.startswith(prefix):
            word = word[len(prefix) :]
            break

    for suffix in suffixes:
        if word.endswith(suffix):
            word = word[: -len(suffix)]
            break

    return word


class MemoryMap:
    """The representation of a memory."""

    def __init__(self, hex_files: Sequence[Path], byteorder: str = "little"):
        """
        :param hex_files: An iterable of hex files paths used to fill the memory map.
        :param byteorder: Byte order to use when retrieving a value from an address range.
        :raises ValueError: If the 'byteorder' is unsupported.
        """
        self._intelhex = IntelHex()
        self._intelhex.padding = None
        for file in hex_files:
            self._intelhex.loadhex(file)

        # Test for supported byteorder
        int.from_bytes((0,), byteorder)
        self._byteorder = byteorder

    def __iter__(self):
        yield from self._intelhex.addresses()

    def __getitem__(self, key: Union[int, range]):
        if isinstance(key, int):
            return self._intelhex[key]

        raw = [self._intelhex[x] for x in key]
        mask = int.from_bytes((0 if x is None else 0xFF for x in raw), self._byteorder)
        value = int.from_bytes((0 if x is None else x for x in raw), self._byteorder)

        return value, mask


def parse_toml(*tomls: Sequence[Union[str, Path]]) -> Record:
    """
    Parse register records extracted from one or more TOML configurations.
    Later configurations override earlier ones.

    :param tomls: Sequence of TOML configuration files with records to load

    :return: Records of register content from TOML configuration files.
    """

    toml_record = Record()

    for config in tomls:
        with Path(config).open("rb") as f_toml:
            toml_record.update(tomli.load(f_toml))

    log_vrb("")
    log_vrb(f"========= TOML Configurations =========\n")
    log_vrb(pformat(toml_record.as_hex()))
    log_vrb("")

    return toml_record


def field_as_toml(field: Field, **kwargs) -> str:
    """
    TOML representation of a register bitfield.

    :kwarg comments: Append comment strings to the field assignment output

    :return: TOML string representation of the field contents.
    """
    comments = kwargs.get("comments")

    reverse_enums: Dict[int, str] = {value: name for name, value in field.enums.items()}

    if field.content in reverse_enums:
        value = f'"{reverse_enums[field.content]}"'
        comment = ", ".join([enum for _value, enum in sorted(reverse_enums.items())])
        default = reverse_enums[field.reset_content]

    elif field.content > TOML_FIELD_AS_HEX_WHEN_LARGER_THAN:
        value = hex(field.content)
        if isinstance(field.allowed_values, range):
            comment = f"0..0x{field.allowed_values.stop - 1:x}"
        else:
            comment = ", ".join([f"{v} (0x{v:x})" for v in field.allowed_values])
        default = hex(field.reset_content)

    else:
        value = str(field.content)
        if isinstance(field.allowed_values, range):
            comment = f"0..{field.allowed_values.stop - 1}"
        else:
            comment = ", ".join([str(v) for v in field.allowed_values])
        default = str(field.reset_content)

    assignment = f"{field.name} = {value}"

    if comments:
        assignment_with_comment = (
            f"{assignment:<{TOML_FIELD_COMMENT_START_AT_COLUMN}}# {comment}"
        )

        return (
            f"{assignment_with_comment:<{TOML_FIELD_DEFAULT_VALUE_START_AT}}"
            f" Reset: {default}"
        )

    return assignment


def register_as_toml(register: Register, **kwargs) -> str:
    """
    TOML representation of a peripheral register.

    :kwarg reset_values: Include unmodified Fields as their reset value
    :kwarg comments: Allow field comment strings in the TOML
    :kwarg force_32_bit_fields: List of Fields that should be given as 32-bit integers, regardless
        of bit width (for example, addresses).

    :return: TOML string representation of the register contents.
    """
    comments = kwargs.get("comments")
    reset_values = kwargs.get("reset_values")
    force_32_bit_fields = kwargs.get("force_32_bit_fields")

    return "\n".join(
        [
            field_as_toml(
                field, comments=comments, force_32_bit_fields=force_32_bit_fields
            )
            for field in register.values()
            if (reset_values or field.modified)
        ]
    )


def generate_toml(out_path: Path, peripheral: LogicalPeripheral, **kwargs):
    """
    Write the peripheral representation to a TOML file.

    This is written without the use of 3rd party libraries, for three reasons:
        1) Most TOML libraries focus on reading, rather than writing.
        2) The libraries that do support writing, do not always handle comments.
        3) Writing TOML is simple.

    :param out_path: Path to write TOML file to.
    :param peripheral: Peripheral to map register content into TOML

    :kwarg reset_values: Include register bit reset values along with modified values.
    :kwarg comments: Include register field comments in the output configuration
    """

    reset_values = kwargs.get("reset_values")
    comments = kwargs.get("comments")

    peripheral_toml: List[str] = []

    for _address, register in sorted(peripheral.address_to_register.items()):
        if reset_values or register.modified:
            register_toml = register_as_toml(
                register,
                reset_values=reset_values,
                comments=comments,
            )
            peripheral_toml.append(
                f"[{peripheral.path_to_record_name[register.path]}]\n{register_toml}"
            )

    file_content = "\n\n".join(peripheral_toml)
    trailing_newline = "\n" if len(file_content) != 0 else ""

    with open(out_path, "w", encoding="utf-8") as file:
        file.write(file_content + trailing_newline)


def generate_hex(out_path: Path, peripheral: LogicalPeripheral, **kwargs):
    """
    Write the peripheral representation to a HEX file.

    :param out_path: Path to write HEX file to.
    :param peripheral: Peripheral to map register content into hex

    :kwarg reset_values: Include register bit reset values along with modified values.
    :kwarg fill_byte: Byte value used to fill unused memory in the peripheral space.
    :kwarg fill: Method for writing fill byte to memory.
    """

    reset_values = kwargs.get("reset_values")
    fill_byte = kwargs.get("fill_byte")
    fill_type = kwargs.get("fill")

    raw_memory: Dict[int, int] = dict(
        peripheral.memory_map if reset_values else peripheral.written_memory_map
    )

    if fill_byte is not None:
        if fill_type == "unmodified":
            # Write the fill byte to all addresses that correspond to a register but that hasn't
            # been modified
            unmodified_addresses = (
                peripheral.memory_map.keys() - peripheral.written_memory_map.keys()
            )
            raw_memory.update({address: fill_byte for address in unmodified_addresses})

        elif fill_type == "all":
            # Write the fill byte to every unmodified address in the peripheral's address range,
            # even those that don't correspond to a register
            gap_addresses = (
                set(chain.from_iterable(peripheral.address_ranges)) - raw_memory.keys()
            )
            raw_memory.update({address: fill_byte for address in gap_addresses})

    ih = IntelHex(raw_memory)
    ih.write_hex_file(out_path)



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

import enum
import importlib.resources
import pickle
from pathlib import Path
from pprint import pformat
from typing import NamedTuple, Optional

import click
import svd
import tomli
from devicetree import edtlib

from . import __version__, bicr, core, uicr
from .common import Product, Record, config_log, log_dbg, log_dev, log_vrb

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
_VERBOSITY = 0


class FormatExtension(enum.Enum):
    """Enumeration of supported formats and their file extension."""

    TOML = ".toml"
    HEX = ".hex"


class Output(NamedTuple):
    start_address: int
    end_address: int
    path: Path


class HexIntParamType(click.ParamType):
    """
    Custom parameter type for hexadecimal integers. This allows click to automatically verify
    hexadecimal parameters when given from the commandline.
    """

    name = "hex_integer"

    def convert(self, value, param, ctx):
        """Overridden method."""

        if isinstance(value, int):
            return value

        try:
            return int(value, 16)
        except ValueError:
            self.fail(f"{value!r} must be in hexadecimal format", param, ctx)


HEX_INT = HexIntParamType()


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Print verbose output such as debug information.",
)
@click.option(
    "-c",
    "--config-file",
    type=click.File("rb", encoding="utf-8"),
    help=(
        "A TOML file containing command options. "
        "The file should follow the schema: <param_n> = <value_of_param_n>. "
        "Options given at the command line take precedence over the values contained in the "
        "configuration file."
    ),
)
@click.pass_context
def cli(ctx={}, verbose=False, config_file=None):
    """
    Generate memory-mapped binary files of peripheral register content, based on
    System View Description (SVD) files and configuration files.

    Use the -h or --help option with each command to see more information about
    them and their individual options/arguments.
    \f
    """
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose

    config_log(verbosity=verbose)

    if config_file is not None:
        try:
            command = ctx.invoked_subcommand
            command_parameters = tomli.load(config_file)
            ctx.default_map = {command: command_parameters}
        except tomli.TOMLDecodeError as exc:
            raise click.BadOptionUsage(
                "config_file",
                f"Failed to parse the '{config_file.name}' as toml.",
                ctx=ctx,
            ) from exc


@cli.command(short_help="Generate hex from a TOML and SVD")
@click.option(
    "-i",
    "--input-config",
    multiple=True,
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    help=(
        "Input TOML configuration. This can be repeated multiple times to combine "
        "multiple configuration files into a single output."
    ),
)
@click.option(
    "-d",
    "--dts-file",
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    help=(
        "Compiled DTS configuration. If given, the script will parse the file for nodes that it "
        "recognizes that are relevant to the given peripheral. If given alongside a "
        "TOML configuration, the TOML contents will be applied after parsing this DTS file."
    ),
)
@click.option(
    "-H",
    "--hex-file",
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    multiple=True,
    help=(
        "HEX file. If given, the script will use it to initialize the values of "
        "the peripheral. Multiple HEX files can be given, and they will be applied in "
        "the order as they are given. Files passed as an input for this option are applied before"
        "any other inputs."
    ),
)
@click.option(
    "-o",
    "--output-file",
    required=True,
    multiple=True,
    type=click.Path(dir_okay=False, exists=False, path_type=Path),
    help=(
        "Generated output file. This can be repeated multiple times to specify "
        "multiple outputs of different formats. Supported formats are: "
        f"{[x.value for x in FormatExtension]}"
    ),
)
@click.option(
    "-p",
    "--peripheral",
    required=True,
    help="Name of the peripheral to be configured.",
)
@click.option(
    "-s",
    "--svd-file",
    required=True,
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    help=("Path to the device SVD file."),
)
@click.option(
    "-D",
    "--bindings-dir",
    multiple=True,
    type=click.Path(dir_okay=True, exists=True, path_type=Path),
    help="Path to a directory to search for devicetree bindings",
)
@click.option(
    "-b",
    "--base-address",
    type=HEX_INT,
    help="Specific base address to use for the peripheral, given in hexadecimal.",
)
@click.option(
    "-f",
    "--fill-byte",
    type=HEX_INT,
    help=(
        "Fill unconfigured memory space within the peripheral memory boundary with this "
        "byte value. This value may or may not be applied based on other options; see --fill for "
        "more information about how this is applied."
    ),
)
@click.option(
    "-F",
    "--fill",
    type=click.Choice(["unmodified", "all"], case_sensitive=False),
    default="unmodified",
    help=(
        "Apply the fill byte given from --fill to unmodified registers within the peripheral "
        "space, or to all unmodified bytes within the entire memory region of the peripheral "
        "boundary. CAUTION: Applying to the entire memory region could cause unintentional "
        "overwriting of other peripheral register values that exist within that space. "
        "Use 'all' only when certain that the peripheral memory space is not shared. "
        "This option is ignored if no --fill-byte is given, or if paired with --reset-values. "
        "TOML outputs will not reflect filled memory from this strategy."
    ),
)
@click.option(
    "-u",
    "--unconstrained",
    multiple=True,
    help=(
        "A register that should be unconstrained, in which it will accept any value "
        "that can fit inside of its fields. An optional field specifier may be added "
        "by appending a dot '.' and the name of the field to unconstrain, which will "
        "leave all other fields in that register untouched. Can be given multiple "
        "times for multiple unconstrained registers."
    ),
)
@click.option(
    "-m",
    "--modified-only",
    is_flag=True,
    help=(
        "Only output values for the registers that have been modified. "
        "With this flag, only registers that have been assigned a different value than their "
        "reset value are written to the output files. This is useful for updating only specific "
        "registers in a peripheral."
    ),
)
@click.option(
    "--comments",
    is_flag=True,
    help=("Print register field comments in generated TOML configurations."),
)
@click.pass_context
def generate(
    ctx,
    input_config,
    dts_file,
    hex_file,
    output_file,
    peripheral,
    svd_file,
    bindings_dir,
    base_address,
    fill_byte,
    fill,
    unconstrained,
    modified_only,
    comments,
):
    """
    Generate binary and/or TOML file(s) for the given peripheral from an SVD and mix of TOML or
    devicetree (DTS) configurations.

    The SVD file provides the foundational definition of what the peripheral looks like.
    Based on this definition, the registers from the configuration will be mapped accordingly into
    their relevant position in memory.
    By default this peripheral contains register contents that with reset values, which are then
    modified by register records contained within the given configurations.

    Passing a compiled DTS file into the tool, such as the zephyr.dts build artifact from Zephyr
    compilations, maps property information from devicetree for nodes to their relevant register
    and bitfield in the associated peripheral.

    TOML configurations are a definition of explicit registers and bitfields that contain
    specific values.
    Specifying more than one TOML configuration will combine the various register configurations
    into a single map to be applied together.
    When paired with a DTS configuration, the TOML configurations will be applied after the parsed
    devicetree contents are interpreted into UICR representation.

    For example, on nRF54H20:

        \b
        nrf-regtool generate -p UICR \\
        -d _build/zephyr/zephyr.dts \\
        -D nrf/dts/bindings \\
        -D modules/lib/nrf-regtool/dts/bindings \\
        -D modules/ic-next/dts/bindings \\
        -D zephyr/dts/bindings \\
        -s modules/hal/nordic_haltium/nrfx/mdk/nrf54h20_application.svd \\
        -o uicr.hex

    Or for a TOML configuration:

        \b
        nrf-regtool generate -p UICR \\
        -i uicr.toml \\
        -s modules/hal/nordic_haltium/nrfx/mdk/nrf54h20_application.svd \\
        -o uicr.hex

    This command accepts multiple input and output files by repeating the related option with a
    different path. This can be useful for combining multiple configuration files together into
    a single output unit, which can be in the form of a hex file or TOML configuration of the
    combined inputs.

    For example:

        \b
        nrf-regtool generate -p UICR \\
        -d _build/zephyr/zephyr.dts \\
        -D nrf/dts/bindings \\
        -D modules/lib/nrf-regtool/dts/bindings \\
        -D modules/ic-next/dts/bindings \\
        -D zephyr/dts/bindings \\
        -s modules/hal/nordic_haltium/nrfx/mdk/nrf54h20_application.svd \\
        -i register_overrides.toml \\
        -o uicr.hex -o uicr_records.toml
    \f
    """

    log_dev(
        pformat(
            {
                "input_config": input_config,
                "dts_file": dts_file,
                "hex_file": hex_file,
                "output_file": output_file,
                "peripheral_name": peripheral,
                "svd_file": svd_file,
                "fill": fill,
                "fill_byte": fill_byte,
                "comments": comments,
                "base_address": base_address,
                "unconstrained": unconstrained,
                "modified_only": modified_only,
            }
        )
    )

    periph_upper = peripheral.upper()

    if dts_file:
        if periph_upper != "BICR":
            raise click.BadParameter(
                f"Generating from DTS is only supported for peripheral BICR",
                ctx=ctx,
            )
        if not bindings_dir:
            raise click.BadParameter(
                f"DTS parsing requires one or more devicetree bindings directories to be specified",
                ctx=ctx,
            )

    if (
        periph_upper in core.NVM_PERIPHERALS
        and fill_byte is not None
        and fill_byte != core.NVM_FILL_BYTE
    ):
        raise click.BadParameter(
            f"Unused UICR memory should only be filled by {hex(core.NVM_FILL_BYTE)!r}; got {hex(fill_byte)!r}",
            ctx=ctx,
        )
    elif base_address is not None and base_address >= 0xFFFFFFFF:
        raise click.BadParameter(
            f"Invalid base address for peripheral {peripheral!r}: {hex(base_address)}",
            ctx=ctx,
        )

    unconstrained_registers = list(
        set(unconstrained) | set(core.REQUIRED_UNCONSTRAINED.get(periph_upper, []))
    )
    if unconstrained_registers:
        log_dbg(f"Unconstrained registers: {pformat(unconstrained_registers)}\n")

    base_address_dts: Optional[int] = None
    peripheral_records = Record()

    log_vrb(f"\n############### Generate {periph_upper} ###############\n")

    if dts_file:
        # Only BICR is supported as checked above
        devicetree = edtlib.EDT(dts_file, bindings_dir)
        base_address_dts, record = bicr.from_dts(devicetree)
        peripheral_records.update(record)

    if input_config:
        peripheral_records.update(core.parse_toml(*input_config))

    if not peripheral_records:
        log_vrb(f"No records compiled for {periph_upper}")
    else:
        log_dbg("")
        log_dbg(
            f"Compiled {periph_upper} records from all inputs:\n{pformat(peripheral_records.as_hex())}\n\n"
        )

    peripheral_base_address: Optional[int]
    if base_address is not None:
        peripheral_base_address = base_address
    elif base_address_dts is not None:
        peripheral_base_address = base_address_dts
    else:
        # Take the base address from the SVD file
        peripheral_base_address = None

    mem_map = core.MemoryMap(hex_file) if hex_file else None

    device = svd.parse(
        svd_file,
        options=svd.Options(
            parent_relative_cluster_address=True,
        ),
    )
    peripheral = core.LogicalPeripheral(
        name=periph_upper,
        device=device,
        record=peripheral_records,
        unconstrained=unconstrained_registers,
        base_address=peripheral_base_address,
        mem_map=mem_map,
    )

    # If DTS was provided, we have a much nicer debug print before this.
    # Otherwise, the raw registers is the best we can show.
    raw_log_fn = log_dbg if dts_file else log_vrb
    raw_log_fn(f"\nRegisters changed from the reset value:\n{peripheral!s}\n\n")

    log_dev(
        f"Registers to be output:\n{peripheral.to_str(lambda r: r.written or not modified_only)}"
    )

    for out_file in [x for x in output_file]:
        if out_file.suffix == FormatExtension.TOML.value:
            core.generate_toml(
                out_file, peripheral, reset_values=not modified_only, comments=comments
            )
        elif out_file.suffix == FormatExtension.HEX.value:
            core.generate_hex(
                out_file,
                peripheral,
                reset_values=not modified_only,
                fill_byte=fill_byte,
                fill=fill,
            )
        else:
            raise click.BadOptionUsage(
                "output_file",
                f"File extension {out_file.suffix!r} is not a supported format; "
                f"must be one of {[x.value for x in FormatExtension]}",
                ctx=ctx,
            )

    log_vrb(f"\n############### End {periph_upper} ###############\n")


@cli.command(short_help="Compile UICR from a DTS")
@click.option(
    "-e",
    "--edt-pickle-file",
    required=True,
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    help=(
        "Pickled edtlib.EDT object with parsed devicetree contents, "
        "containing information that is relevant for a domain's UICR."
    ),
)
@click.option(
    "-o",
    "--output-file",
    required=True,
    multiple=True,
    type=click.Path(dir_okay=False, exists=False, path_type=Path),
    help=(
        "Generated output file in HEX format. "
        "Multiple output paths can be given by repeating this option."
    ),
)
@click.option(
    "-P",
    "--product-name",
    required=True,
    help=(
        "Name of the Nordic SoC or SiP for which to compile the UICR. "
        f"Supported values are: {', '.join(x.name for x in Product)}"
    ),
)
@click.pass_context
def uicr_compile(
    ctx,
    edt_pickle_file,
    output_file,
    product_name,
):
    """
    User Information Configuration Registers (UICR), in the context of certain
    Nordic SoCs, are used by Secure Domain Firmware (SDFW) to configure system
    resources, like memory and peripherals. When used in singular, a UICR refers
    to a complete set of registers for one domain.

    This command compiles a UICR binary blob from devicetree input, which should
    indicate the system resource configuration required for a single domain.

    For example, on nRF54H20:

        \b
        PYTHONPATH=zephyr/scripts/dts/python-devicetree/src \\
        nrf-regtool uicr-compile \\
        -e _build/zephyr/edt.pickle \\
        -o uicr.hex \\
        -P NRF54H20

    This command is intended for use as part of the Zephyr RTOS application
    development workflow. nrf-regtool depends on edtlib, a devicetree parsing
    library provided by the Zephyr Project, and the input file would usually be
    the edt.pickle artifact from a Zephyr build system. Using the pickle input
    requires a matching revision of edtlib from the Zephyr repository.
    \f
    """

    try:
        product = Product(product_name)
    except ValueError:
        raise click.BadOptionUsage(
            "product_name",
            f"Unknown product name {product_name}; must be one of {[x.name for x in Product]}",
            ctx=ctx,
        )

    log_vrb(f"\n############### Compile UICR ###############\n")

    with edt_pickle_file.open("rb") as file:
        devicetree = pickle.load(file)

    base_address, record = uicr.from_dts(devicetree, product)

    if not record:
        log_vrb(f"No records compiled for UICR")
    else:
        log_dbg(
            f"\nCompiled UICR records from all inputs:\n{pformat(record.as_hex())}\n\n"
        )

    with importlib.resources.path("nrfregtool.resources", "uicr.svd") as svd_file:
        device = svd.parse(
            svd_file,
            options=svd.Options(
                parent_relative_cluster_address=True,
            ),
        )

    peripheral = core.LogicalPeripheral(
        name="UICR",
        device=device,
        record=record,
        base_address=base_address,
    )

    log_dbg(f"\nRegisters changed from the reset value:\n{peripheral!s}\n\n")
    log_dev(f"Registers to be output:\n{peripheral.to_str()}")

    for out_file in output_file:
        core.generate_hex(out_file, peripheral, reset_values=True)

    log_vrb(f"\n############### End UICR ###############\n")



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

import importlib.metadata


PACKAGE_NAME = "nrf-regtool"

__version__ = importlib.metadata.version(PACKAGE_NAME)

from .common import (
    AddressRegion,
    AddressOffset,
    DomainID,
    OwnerID,
    ProcessorID,
    Product,
    Record,
    config_log,
    log_vrb,
    log_dbg,
    log_err,
)
from .core import (
    LogicalPeripheral,
    parse_toml,
    field_as_toml,
    register_as_toml,
    generate_toml,
    generate_hex,
    split_register_path,
)
from .parsed_dt import (
    ParsedDTNode,
    ProcessorInfo,
    dt_processor_info,
)
from .uicr import (
    GpioPin,
    IpcMapConfig,
    SourceMapIndex,
    SinkMapIndex,
    NodeType,
    ResourceCompatible,
    InterpretedUICR,
    extract_memory_configs,
    extract_periph_configs,
    extract_gpio_configs,
    extract_gpiote_configs,
    extract_grtc_configs,
    extract_dppic_configs,
    extract_ipct_configs,
    extract_vtor_configs,
    extract_ptrextuicr_config,
)
from .bicr import (
    InterpretedBICR,
    PowerConfig,
    extract_power_configs,
)



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from .cli import cli


def package_main() -> None:
    cli(prog_name="nrf-regtool")


if __name__ == "__main__":
    package_main()
